"""Float and currency conversion for data feeds."""
import enum

import pandas as pd


class CurrencyConversion(enum.Enum):
    """How prices in the OHLCV output are handled."""

    #: Everything is converted to US dollars
    us_dollar = "us_dollar"

    #: Everything is kept in its native crypto token.
    #:
    #: Nothing is done for the prices
    crypto_quote_token = "quote_token"


def convert_to_float(df: pd.DataFrame, conversion: CurrencyConversion) -> pd.DataFrame:
    """Convert raw transaction values ot USD before resample.

    - This will create a copy of the dataframe.

    - The resulting DataFrame is going to lose columns
      that are not on its retained list

    - Columns are going to be converted to floats

    :param df:
        Raw trades dataframe

    :param conversion:
        Do we want the resulting dataframe to be converted
        using the given exchange rate

    :return:
        Dataframe with floating point USD pricing.


        The resulting dataframe is still indexed by block.
    """

    df2 = pd.DataFrame(index=df.index)
    df2["pair"] = df["pair"]
    df2["timestamp"] = df["timestamp"]
    df2["block_number"] = df["block_number"]
    df2["exchange_rate"] = df["exchange_rate"].astype(float)
    if conversion == CurrencyConversion.us_dollar:
        df2["price"] = df["price"].astype(float) * df2["exchange_rate"]
        df2["amount"] = df["amount"].astype(float) * df2["exchange_rate"]
    else:
        df2["price"] = df["price"].astype(float)
        df2["amount"] = df["amount"].astype(float)
    return df2



"""Liquidity data feed manipulation.

For more information about liquidity in automatic market making pools see :term:`AMM`
and :term:`XY liquidity model`.
"""

import datetime
from dataclasses import dataclass
from typing import List, Optional, Iterable, Tuple, Dict

import pandas as pd
import pyarrow as pa
from dataclasses_json import dataclass_json
from pandas.core.groupby import GroupBy, DataFrameGroupBy

from tradingstrategy.timebucket import TimeBucket
from tradingstrategy.types import UNIXTimestamp, USDollarAmount, BlockNumber, PrimaryKey
from tradingstrategy.utils.groupeduniverse import PairGroupedUniverse


# Preconstructed pd.Tiemdelta for optimisation
_ZERO_TIMEDELTA = pd.Timedelta(0)


class LiquidityDataUnavailable(Exception):
    """We tried to look up liquidity info for a trading pair, but count not find a sample."""


@dataclass_json
@dataclass
class XYLiquidity:
    """Data structure that presents liquidity status in bonding curve pool.

    This data structure is for naive x*y=k :term:`AMM` pool.
    Liquidity is not the part of the normal :term:`technical analysis`,
    so the dataset server has separate datasets for it.

    Liquidity is expressed as US dollar value of the :term:`quote token` of the pool.
    For example if the pool is 50 $FOO in reserve0 and 50 $USDC in reserve1, the
    liquidity of the pool would be expressed as 50 USD.

    Liquidity events, like :term:`candles <candle>`, have open, high, low and close values,
    depending on which time of the candle they were sampled.
    """

    #: Primary key to identity the trading pair
    #: Use pair universe to map this to chain id and a smart contract address
    pair_id: PrimaryKey

    #: Open timestamp for this time bucket.
    timestamp: UNIXTimestamp

    #: USD exchange rate of the quote token used to
    #: convert to dollar amounts in this time bucket.
    #:
    #: Note that currently any USD stablecoin (USDC, DAI) is
    #: assumed to be 1:1 and the candle server cannot
    #: handle exchange rate difference among stablecoins.
    #:
    #: The rate is taken at the beginning of the 1 minute time bucket.
    #: For other time buckets, the exchange rate is the simple average
    #: for the duration of the bucket.
    exchange_rate: float

    #: Liquidity absolute values in the pool in different time points.
    #: Note - for minute candles - if the candle contains only one event (mint, burn, sync)
    #: the open liquidity value is the value AFTER this event.
    #: The dataset server does not track the closing value of the previous liquidity event.
    #: This applies for minute candles only.
    open: USDollarAmount

    #: Liquidity absolute values in the pool in different time points
    close: USDollarAmount

    #: Liquidity absolute values in the pool in different time points
    high: USDollarAmount

    #: Liquidity absolute values in the pool in different time points
    low: USDollarAmount

    #: Number of liquidity supplied events for pool
    adds: int

    #: Number of liquidity removed events for the pool
    removes: int

    #: Number of total events affecting liquidity during the time window.
    #: This is adds, removes AND swaps AND sync().
    syncs: int

    #: How much new liquidity was supplied, in the terms of the quote token converted to US dollar
    add_volume: USDollarAmount

    #: How much new liquidity was removed, in the terms of the quote token converted to US dollar
    add_volume: USDollarAmount

    #: Blockchain tracking information
    start_block: BlockNumber

    #: Blockchain tracking information
    end_block: BlockNumber

    def __repr__(self):
        human_timestamp = datetime.datetime.utcfromtimestamp(self.timestamp)
        return f"@{human_timestamp} O:{self.open} H:{self.high} L:{self.low} C:{self.close} V:{self.volume} A:{self.adds} R:{self.removes} SB:{self.start_block} EB:{self.end_block}"

    @classmethod
    def to_pyarrow_schema(cls, small_candles=False) -> pa.Schema:
        """Construct schema for writing Parquet filess for these candles.

        :param small_candles: Use even smaller word sizes for frequent (1m) candles.
        """
        schema = pa.schema([
            ("pair_id", pa.uint32()),
            ("timestamp", pa.timestamp("s")),
            ("exchange_rate", pa.float32()),
            ("open", pa.float32()),
            ("close", pa.float32()),
            ("high", pa.float32()),
            ("low", pa.float32()),
            ("adds", pa.uint16() if small_candles else pa.uint32()),
            ("removes", pa.uint16() if small_candles else pa.uint32()),
            ("syncs", pa.uint16() if small_candles else pa.uint32()),
            ("add_volume", pa.float32()),
            ("remove_volume", pa.float32()),
            ("start_block", pa.uint32()),   # Should we good for 4B blocks
            ("end_block", pa.uint32()),
        ])
        return schema

    @classmethod
    def to_dataframe(cls) -> pd.DataFrame:
        """Return emptry Pandas dataframe presenting liquidity sample."""

        # TODO: Does not match the spec 1:1 - but only used as empty in tests
        fields = dict([
            ("pair_id", "int"),
            ("timestamp", "datetime64[s]"),
            ("exchange_rate", "float"),
            ("open", "float"),
            ("close", "float"),
            ("high", "float"),
            ("low", "float"),
            ("buys", "float"),
            ("sells", "float"),
            ("add_volume", "float"),
            ("remove_volume", "float"),
            ("start_block", "float"),
            ("end_block", "float"),
        ])
        df = pd.DataFrame(columns=fields.keys())
        return df.astype(fields)


@dataclass_json
@dataclass
class LiquidityResult:
    """Server-reply for live queried liquidity data."""

    #: A bunch of candles.
    #: Candles are unordered and subject to client side sorting.
    #: Multiple pairs and chains may be present in candles.
    liquidity_events: List[XYLiquidity]

    def sort_by_timestamp(self):
        """In-place sorting of candles by their timestamp."""
        self.candles.sort(key=lambda c: c.timestamp)


class GroupedLiquidityUniverse(PairGroupedUniverse):
    """A universe where each trading pair has its own liquidity data feed.

    This is helper class to create foundation for multi pair strategies.

    For the data logistics purposes, all candles are lumped together in single columnar data blobs.
    However, it rarely makes sense to execute operations over different trading pairs.
    :py:class`GroupedLiquidityUniverse` creates trading pair id -> liquidity sample data grouping out from
    raw liquidity sample.
    """

    def __init__(self,
                 df: pd.DataFrame,
                 time_bucket=TimeBucket.d1,
                 timestamp_column="timestamp",
                 index_automatically=True):
        super().__init__(df, time_bucket, timestamp_column, index_automatically)

        # For fast liquidity lookup
        self.indexer_cache: Dict[PrimaryKey, int] = {}

    def get_liquidity_samples_by_pair(self, pair_id: PrimaryKey) -> Optional[pd.DataFrame]:
        """Get samples for a single pair.

        If the pair does not exist return `None`.
        """
        try:
            return self.get_samples_by_pair(pair_id)
        except KeyError:
            return None

    def get_liquidity_with_tolerance(self,
                          pair_id: PrimaryKey,
                          when: pd.Timestamp,
                          tolerance: pd.Timedelta,
                          kind="close") -> Tuple[USDollarAmount, pd.Timedelta]:
        """Get the available liquidity for a trading pair at a specific time point/

        The liquidity is defined as one-sided as in :term:`XY liquidity model`.

        Example:

        .. code-block:: python

            liquidity_amount, last_trade_delay = liquidity_universe.get_liquidity_with_tolerance(
            sushi_usdt.pair_id,
                pd.Timestamp("2021-12-31"),
                tolerance=pd.Timedelta("1y"),
            )
            assert liquidity_amount == pytest.approx(2292.4517)
            assert last_trade_delay == pd.Timedelta('4 days 00:00:00')

        :param pair_id:
            Trading pair id

        :param when:
            Timestamp to query

        :param kind:
            One of OHLC data points: "open", "close", "low", "high"

        :param tolerance:
            If there is no liquidity sample available at the exact timepoint,
            look to the past to the get the nearest sample.
            For example if candle time interval is 5 minutes and look_back_timeframes is 10,
            then accept a candle that is maximum of 50 minutes before the timepoint.

        :return:
            Return (price, delay) tuple.

            We always return a price. In the error cases an exception is raised.
            The delay is the timedelta between the wanted timestamp
            and the actual timestamp of the candle.

            Candles are always timestamped by their opening.

        :raise LiquidityDataUnavailable:

            There were no samples available with the given condition.

            This can happen when

            - There has not been a single trade yet

            - There hasn't been any trades since `tolerance`
              time window

        """

        assert kind in ("open", "close", "high", "low"), f"Got kind: {kind}"

        last_allowed_timestamp = when - tolerance

        try:
            candles_per_pair = self.get_samples_by_pair(pair_id)
        except KeyError as e:
            raise LiquidityDataUnavailable(f"Pair {pair_id} does not contain any liquidity samples at all") from e

        samples_per_kind = candles_per_pair[kind]

        # https://pandas.pydata.org/docs/reference/api/pandas.Index.get_indexer.html
        # https://stackoverflow.com/questions/71027193/datetimeindex-get-loc-is-deprecated
        indexer = candles_per_pair.index.get_indexer([when], method="pad")

        if indexer[0] != -1:
            discrete_ts_idx = indexer[0]
        else:
            first_sample = candles_per_pair.index[0]
            raise LiquidityDataUnavailable(f"Pair {pair_id} does not have data older data than {first_sample}. Tried to look up {when}.")

        sample_timestamp = candles_per_pair.index[discrete_ts_idx]

        latest_value = candles_per_pair.iloc[discrete_ts_idx][kind]

        distance = when - sample_timestamp
        assert distance >= _ZERO_TIMEDELTA, f"Somehow we managed to get a timestamp {sample_timestamp} that is newer than asked {when}"

        if sample_timestamp >= last_allowed_timestamp:
            # Return the chosen price column of the sample
            return latest_value, distance

        # Try to be helpful with the errors here,
        # so one does not need to open ipdb to inspect faulty data
        try:
            first_sample = candles_per_pair.iloc[0]
            second_sample = candles_per_pair.iloc[1]
            last_sample = candles_per_pair.iloc[-1]
        except KeyError:
            raise LiquidityDataUnavailable(
                f"Could not find any liquidity samples for pair {pair_id}, value kind '{kind}', between {when} - {last_allowed_timestamp}\n"
                f"Could not figure out existing data range. Has {len(samples_per_kind)} samples."
            )

        raise LiquidityDataUnavailable(
            f"Could not find any liquidity samples for pair {pair_id}, value kind '{kind}', between {when} - {last_allowed_timestamp}\n"
            f"The pair has {len(samples_per_kind)} candles between {first_sample['timestamp']} - {last_sample['timestamp']}\n"
            f"Sample interval is {second_sample['timestamp'] - first_sample['timestamp']}\n"
            f"\n"
            f"This is usually due to sparse candle data - trades have not been made or the blockchain was halted during the price look-up period.\n"
            f"Try to increase look back perid in your code."
            )

    def get_closest_liquidity(self, pair_id: PrimaryKey, when: pd.Timestamp, kind="open", look_back_time_frames=5) -> USDollarAmount:
        """Get the available liuqidity for a trading pair at a specific timepoint or some candles before the timepoint.

        The liquidity is defined as one-sided as in :term:`XY liquidity model`.

        .. warning::

                This is an early alpha method and has been deprecated.
                Please use  :py:meth:`get_liquidity_with_tolerance` instead.

        :param pair_id: Traing pair id
        :param when: Timestamp to query
        :param kind: One of liquidity samples: "open", "close", "low", "high"
        :param look_back_timeframes: If there is no liquidity sample available at the exact timepoint,
            look to the past to the get the nearest sample
        :return: We always return
        :raise LiquidityDataUnavailable: There was no liquidity sample available
        """

        assert kind in ("open", "close", "high", "low"), f"Got kind: {kind}"

        start_when = when
        samples_per_pair = self.get_liquidity_samples_by_pair(pair_id)
        assert samples_per_pair is not None, f"No liquidity data available for pair {pair_id}"

        samples_per_kind = samples_per_pair[kind]
        for attempt in range(look_back_time_frames):
            try:
                sample = samples_per_kind[when]
                return sample
            except KeyError:
                # Go to the previous sample
                when -= self.time_bucket.to_timedelta()

        raise LiquidityDataUnavailable(f"Could not find any liquidity samples for pair {pair_id} between {when} - {start_when}")

    @staticmethod
    def create_empty() -> "GroupedLiquidityUniverse":
        """Create a liquidity universe without any data."""
        return GroupedLiquidityUniverse(df=XYLiquidity.to_dataframe(), index_automatically=False)


class ResampledLiquidityUniverse:
    """Backtesting performance optimised version of liquidity universe.

    - The class is designed to be used with strategies that trade
      a large number of pairs and picks pairs by their liquidity criteria,
      which varies over time

    - Because liquidity information does not need to be as accurate as pricing information,
      and a rough estimate of liquidity is enough for most strategies, we can preconstruct
      an optimised version of liquidity universe

    - This version of liquidity universe is designed for the maximum speed of
      :py:meth:`get_liquidity_fast` with (pair id, timestamp) parameter

    - Some of the speed is achieved by precomputing upsampling liquidity data
      to a longer frequency (day, week)

    - Some of the speed is achieved by using extra indexes and memory

    - Some of the speed is achieved by forward filling any data gaps,
      so we do not need to worry about the sparse data

    The speedup difference between :py:meth`LiquidityUniverse.get_liquidity_with_tolerance`
    and py:meth:`get_liquidity_fast` is more than 10x.
    The practice speed increase for multipair liquidity aware strategy backtest can be
    2x - 3x.

    .. note ::

        The resampling itself will take some time, so optimally
        the resampled data should be stored on-disk cache between different backtest runs

    """

    def __init__(self,
                df: pd.DataFrame,
                column="close",
                resample_period="1D",
                resample_method="min",
        ):
        """Calculate optimised liquidity universe.

        - Only extract the column we are interested in

        - Resample to a smaller  data

        - Resample by the

        .. note::

            For a large number of pairs, resampling will take a long time.
            You should remove any unnecessary pairs in `df` before creating a resampled version.

        :param df:

            Liquidity dataframe.

            You should prepare this by dropping any unwanted pairs.

        :param column:
            Which column value we use for the resampled series.

            Defaults to (daily) close liquidity value.

        :param resample_period:

            Pandas frequency alias string for the new timeframe duration.

            E.g. `1D` for daily.

            Must be a fixed frequency e.g. `30D` instead of `M`.

            See https://stackoverflow.com/a/35339226/315168

        :param resample_method:
            How to we resample the liquidity

        """
        new_df = df[["timestamp", "pair_id"]].copy()
        new_df["value"] = df[[column]]
        new_df.set_index(new_df["timestamp"], inplace=True, drop=True)
        grouped_df = new_df.groupby("pair_id")

        # https://pandas.pydata.org/docs/reference/api/pandas.core.groupby.DataFrameGroupBy.resample.html
        #
        # Important to set origin="epoch" here, or otherwise
        # pd.Timestamp().floor() does not match our range values in get_liquidity_fast()
        #
        resampled_df = grouped_df.resample(resample_period, origin="epoch").agg({"value": resample_method}).ffill()
        self.resample_period = resample_period
        self.df: pd.DataFrame = resampled_df

        self.pair_cache: Dict[PrimaryKey, pd.DataFrame] = {}

    def get_samples_by_pair(self, pair_id: int) -> pd.DataFrame:
        """Access and cache the resampled liquidity data per pair."""
        # https://stackoverflow.com/a/45563615/315168

        if pair_id not in self.pair_cache:
            try:
                self.pair_cache[pair_id] = self.df.xs(pair_id)
            except KeyError as e:
                raise KeyError(f"Could not find pair for pair_id {pair_id}") from e
        return self.pair_cache[pair_id]

    def get_liquidity_fast(self,
                      pair_id: PrimaryKey,
                      when: pd.Timestamp) -> USDollarAmount:
        """Get the available liquidity for a trading pair at a specific time point.

        The liquidity is read from an upsampled buffer.

        The lookup is fast because any timestamp is rounded down to the nearest frequency full timestamp
        and looked up there.

        :param pair_id:
            Trading pair id

        :param when:
            Timestamp to query

        :return:
            Return available liquidty or `0` if no liquidity was seen before the timestamp.
        """
        assert isinstance(when, pd.Timestamp)

        try:
            samples = self.get_samples_by_pair(pair_id)
        except KeyError as e:
            raise LiquidityDataUnavailable(f"No liquidity data for {pair_id}") from e

        try:
            rounded_ts = when.floor(self.resample_period)
            return samples.loc[rounded_ts]["value"]
        except KeyError:
            return 0.0



"""Timeframe helpers."""

from dataclasses import dataclass

import pandas as pd


#: Default chart settings for different timeframes.
#:
#: A sensible default of a history for different candle timeframes.
#:
#: Timeframe -> the default history lookback duration mappings.
#: Have around 100 candles per timeframe with the default zoom level.
#:
DEFAULT_DISPLAY_WINDOW_FOR_TIMEFRAME = {
    "1min": pd.Timedelta("2h"),
    "5min": pd.Timedelta("6h"),
    "1h": pd.Timedelta("3d"),
}


@dataclass
class Timeframe:
    """Describe candle timeframe.

    This structure allows us to pass candle resample
    data around the framework.

    This class is very similar to :py:class:`tradingstrategy.timebucket.TimeBucket`,
    but unlike the bucket it allows any time frames presented, not just predefined choices.

    These is also :py:attr:`offset` which allows to shift the starting point of candles
    around.
    """

    #: Pandas frequency string.
    #:
    #: E.g. `1D` for daily, `1min` for minute.
    #:
    #: See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
    freq: str

    #: Candle shift offset
    #:
    #: E.g. move hourly candles 5 minutes backwards to start at 00:55
    offset: pd.Timedelta = pd.Timedelta(seconds=0)

    def __eq__(self, other):
        """Implemented for set() and dict()"""
        assert isinstance(other, Timeframe)
        return self.freq == other.freq and self.offset == other.offset

    def __hash__(self) -> int:
        """Implemented for set() and dict()"""
        return (self.freq, self.offset).__hash__()

    def __repr__(self) -> str:
        return f"<Timeframe freq:{self.freq} offset:{self.offset}>"

    def round_timestamp_down(self, ts: pd.Timestamp) -> pd.Timestamp:
        """Snap to previous available timedelta.

        Preserve any timezone info on `ts`.

        If we are at the the given exact delta, then do not round, only add offset.

        :param ts:
            Timestamp we want to round

        :return:
            When to wake up from the sleep next time
        """
        return ts.floor(self.freq) + self.offset

    def get_default_chart_display_window(self) -> pd.Timedelta:
        """Get the default candle chart time buffer for this frequency.

        See :py:attr:`DEFAULT_DISPLAY_WINDOW_FOR_TIMEFRAME`.

        :return:
            Pandas frequency string
        """
        window = DEFAULT_DISPLAY_WINDOW_FOR_TIMEFRAME.get(self.freq)
        if window:
            return window
        return pd.Timedelta("24h")



import warnings


def disable_pandas_warnings():
    #   /Users/moo/code/ts/trade-executor/deps/trading-strategy/tradingstrategy/direct_feed/trade_feed.py:338: FutureWarning: The behavior of Timestamp.utcfromtimestamp is deprecated, in a future version will return a timezone-aware Timestamp with UTC timezone. To keep the old behavior, use Timestamp.utcfromtimestamp(ts).tz_localize(None). To get the future behavior, use Timestamp.fromtimestamp(ts, 'UTC')
    #warnings.filterwarnings(action='ignore', category=FutureWarning, message=".*utcfromtimestamp.*")
    warnings.filterwarnings(action='ignore', category=FutureWarning)


"""Analyse the portfolio development over the time.

.. warning ::

    This module is deprecated and replaced by tradeexecutor analysis modules.

- Portfolio situation at the start of the each tick
- Currently held assets
- Net asset value (NAV)
- Asset valuation change
"""

from dataclasses import dataclass, field
from typing import List, Dict, Iterable, Optional, Tuple, Callable

import numpy as np
import pandas as pd
from pandas.io.formats.style import Styler

from tradingstrategy.exchange import ExchangeUniverse
from tradingstrategy.pair import DEXPair, LegacyPairUniverse
from tradingstrategy.types import USDollarAmount, PrimaryKey


@dataclass
class AssetSnapshot:
    """Asset holdings at a specific timepoint."""

    #: How many tokens of the asset we had
    quantity: float

    #: Daily market value of the tokens
    market_value: USDollarAmount

    realised_pnl: USDollarAmount

    unrealised_pnl: USDollarAmount

    total_pnl: USDollarAmount

    # https://docs.python.org/3/library/dataclasses.html#post-init-processing
    def __post_init__(self):
        assert self.quantity > 0, f"Bad quantity {self.quantity}"
        assert self.market_value > 0, f"Bad market value {self.market_value}"
        assert type(self.total_pnl) == float
        assert type(self.realised_pnl) == float
        assert type(self.unrealised_pnl) == float


@dataclass
class PortfolioSnapshot:
    """Represents the portfolio status at the start of the day/candle snapshot"""

    #: A running counter where the first backtest simulated event is tick 1, the next one tick 2
    tick: int

    #: What reserve currencies we have.
    #: E.g. USD: 10_000
    #: Resever currenecies expressed as strings for the backwards compatibiltiy.
    cash_balances: Dict[str, float]

    #: What reserve currencies we have.
    #: E.g. USD: 10_000
    #: Resever currenecies expressed as strings for the backwards compatibiltiy.
    asset_snapshots: Dict[PrimaryKey, AssetSnapshot]

    #: Internal state dump of the algorithm when this trade was made.
    #: This is mostly useful when doing the trade analysis try to understand
    #: why some trades were made.
    #: It also allows you to reconstruct the portfolio state over the time.
    state_details: Optional[Dict] = None

    def get_ordered_assets_stable(self) -> List[Tuple[PrimaryKey, AssetSnapshot]]:
        """Return asset snapshots in a stable order between days.
        """
        assets = [(pair_id, s) for pair_id, s in self.asset_snapshots.items()]
        assets.sort(key=lambda a: a[0])
        return assets

    def get_ordered_assets_by_weight(self) -> List[Tuple[PrimaryKey, AssetSnapshot]]:
        """Return asset snapshots in an order where the heaviest asset is first.
        """
        assets = [(pair_id, s) for pair_id, s in self.asset_snapshots.items()]
        assets.sort(key=lambda a: a[1].market_value, reverse=True)
        return assets


@dataclass
class PortfolioAnalyzer:
    """Represents the portfolio analysis over the backtest period."""

    snapshots: Dict[pd.Timestamp, PortfolioSnapshot]

    def get_max_assets_held_once(self) -> int:
        """Find out what was the max number of assets the strategy was holding at the same time."""
        max_hold = 0
        for s in self.snapshots.values():
            max_hold = max(max_hold, len(s.asset_snapshots))
        return max_hold


def expand_snapshot_to_row(
        exchange_universe: ExchangeUniverse,
        pair_universe: LegacyPairUniverse,
        ts: pd.Timestamp,
        snapshot: PortfolioSnapshot,
        max_assets: int,
        timestamp_format: str) -> dict:
    """Create DataFrame rows from each portfolio snapshot."""

    # timestamp = row.name  # ???
    assert max_assets

    r = {
        # "timestamp": timestamp,
        "Id": snapshot.tick,
        "Holdings at": ts.strftime(timestamp_format),
        "NAV USD": 0,
        "Cash USD": f"{snapshot.cash_balances['USD']:,.0f}",
    }

    assets = snapshot.get_ordered_assets_by_weight()

    # Initialize empty asset columns
    for i in range(max_assets):
        idx = i + 1
        r[f"#{idx} asset"] = pd.NA
        r[f"#{idx} value"] = pd.NA
        r[f"#{idx} weight %"] = pd.NA
        r[f"#{idx} PnL"] = pd.NA
        r[f"#{idx} PnL raw"] = 0

    total_asset_value = 0
    for i, asset_tuple in enumerate(assets):
        idx = i + 1
        pair_id, asset_snapshot = asset_tuple
        pair = pair_universe.get_pair_by_id(pair_id)
        r[f"#{idx} asset"] = pair.base_token_symbol[0:8]  # Cut long ticker names
        r[f"#{idx} value"] = f"{asset_snapshot.market_value:,.0f}"
        r[f"#{idx} PnL"] = f"{asset_snapshot.total_pnl:,.2f}"
        r[f"#{idx} PnL raw"] = asset_snapshot.total_pnl
        assert type(asset_snapshot.total_pnl) == float
        total_asset_value += asset_snapshot.market_value

    for i, asset_tuple in enumerate(assets):
        idx = i + 1
        pair_id, asset_snapshot = asset_tuple
        value = r[f"#{idx} value"]
        if value:
            r[f"#{idx} weight %"] = f"{asset_snapshot.market_value / total_asset_value * 100:.0f}"

    r["NAV USD"] = f"{snapshot.cash_balances['USD'] + total_asset_value:,.2f}"

    return r


def expand_timeline(
        exchange_universe: ExchangeUniverse,
        pair_universe: LegacyPairUniverse,
        analyzer: PortfolioAnalyzer,
        create_html_styles=True,
        vmin=-0.3,
        vmax=0.2,
        timestamp_format="%Y-%m-%d",
) -> pd.DataFrame:
    """Console output for the portfolio development over the time.

    Each row presents the portfolio status at the end of the day/candle.

    The outputted data frame is intented to be human readable and not for programmatic manipulation.

    :param create_html_styles: Create a formatter function that can be applied to hide and recolour columns.

    :param vmax: Trade success % to have the extreme green color.

    :param vmin: The % of lost capital on the trade to have the extreme red color.

    :param timestamp_format: How to format Opened at column, as passed to `strftime()`

    :return: pd.Dataframe rendering the portfolio development over the time
    """

    asset_column_count = analyzer.get_max_assets_held_once()

    raw_output = [expand_snapshot_to_row(exchange_universe, pair_universe, ts, s, asset_column_count, timestamp_format) for ts, s in analyzer.snapshots.items()]

    applied_df = pd.DataFrame(raw_output) # timeline.apply(expander, axis='columns', result_type='expand')

    # Sort portfoli snapshots by backtest tick events
    # https://stackoverflow.com/a/52720936/315168
    applied_df\
        .sort_values(by=['Id'], ascending=[True], inplace=True)

    # Get rid of NaN labels
    # https://stackoverflow.com/a/28390992/315168
    applied_df.fillna('', inplace=True)

    def apply_styles(df: pd.DataFrame):
        # Create a Pandas Styler with multiple styling options applied
        # https://www.geeksforgeeks.org/make-a-gradient-color-mapping-on-a-specified-column-in-pandas/
        # Dynamically color the background of trade outcome coluns # https://pandas.pydata.org/docs/reference/api/pandas.io.formats.style.Styler.background_gradient.html
        styles = df.style.hide_index()
        hidden_columns = []
        asset_colums = []
        for i in range(asset_column_count):
            # Add a background color for each asset column group
            idx = i + 1
            styles = styles.background_gradient(
                axis=0,
                subset=[f"#{idx} asset", f"#{idx} value", f"#{idx} PnL", f"#{idx} weight %"],
                gmap=applied_df[f"#{idx} PnL raw"],
                cmap='RdYlGn',
                vmin=vmin,  # We can only lose 100% of our money on position
                vmax=vmax)  # 50% profit is 21.5 position. Assume this is the max success color we can hit over

            # Add more border between assets
            # https://coderzcolumn.com/tutorials/python/simple-guide-to-style-display-of-pandas-dataframes
            # https://pandas.pydata.org/docs/reference/api/pandas.io.formats.style.Styler.set_table_styles.html
            hidden_columns.append(f"#{idx} PnL raw")
            asset_colums.append(f"#{idx} asset")

        # Build table col styles

        styles_dict = {
            # Don't break timestamp value to multiple lines
            "Holdings at": [{'selector': 'td', 'props': [('white-space', 'nowrap')]}],
        }

        # Format asset column groups
        for col in asset_colums:
            styles_dict[col] = [{'selector': 'td', 'props': [('border-left', '3px solid #888'), ("font-weight", "bold"), ("text-align", "left")]}]

        styles = styles.set_table_styles(styles_dict)
        styles = styles.hide_columns(hidden_columns)
        return styles

    if create_html_styles:
        return applied_df, apply_styles
    else:
        # Format for console
        for i in range(asset_column_count):
            idx = i + 1
            del applied_df[f"#{idx} PnL raw"]
        return applied_df,  None



"""Show histograms how many trades were profitable or not.

.. warning ::

    This module is deprecated and replaced by tradeexecutor analysis modules.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as pyplot
from matplotlib.axes import SubplotBase
from matplotlib.axes._base import _AxesBase
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.container import BarContainer
from matplotlib.figure import Figure



def plot_trade_profit_distribution(df: pd.DataFrame, bins=25, cap_profit=1.1, vmin=-1, vmax=0.5) -> Figure:
    """Create a histogram of won and lost trades based on trade analyzer expanded timeline output.

    See also :py:meth:`tradingstrategy.analysis.tradeanalyizer.expand_timeline`.

    :param df: A DataFrame with a column `PnL % raw`

    :param cap_profit: If profit is higher than this, cap the value so that one off mega profits to not break the diagram
    """

    colormap: LinearSegmentedColormap = pyplot.cm.get_cmap("RdYlGn")
    df = df[["PnL % raw"]].clip(-1, cap_profit)

    # A single successfuk trade might be something like 400% or 4.0
    max_profit_pct = df["PnL % raw"].max()
    min_profit_pct = df["PnL % raw"].min()

    # https://stackoverflow.com/questions/23061657/plot-histogram-with-colors-taken-from-colormap
    Y: np.array
    X: np.array
    # https://numpy.org/doc/stable/reference/generated/numpy.histogram.html
    # Always adjust left X to -100% or trade lost all money
    Y, X = np.histogram(df, bins, range=(min_profit_pct, max_profit_pct))
    x_span = X.max() - X.min()

    C = []
    for x in X:

        # Map x back to the range -100% to 100% - 100% is the max green.
        # Do not directly use the x co-ordinate on the bar diagram for the color
        # as middle is not zero

        # CLip
        profit_clipped = min(vmax, x)
        profit_clipped = max(vmin, profit_clipped)
        green_percent = profit_clipped / (vmax - vmin)
        colormap_adjusted = green_percent/2 + 0.5  # 0 ... 1.0

        c = colormap(colormap_adjusted)
        # print(c, x, profit_pct, X.min(), x_span)
        C.append(c)

    fig: Figure
    ax: _AxesBase
    fig, ax = pyplot.subplots()

    # https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.bar.html
    bar_container: BarContainer = ax.bar(
        x=X[:-1],
        height=Y,
        color=C,
        width=X[1] - X[0])

    xtick_labels = [f"{x:.0%}" for x in ax.get_xticks()]
    ax.set_xlabel("Trade profit %")
    ax.set_ylabel("Number of trades")
    ax.set_xticklabels(xtick_labels)
    ax.set_title("Trade won/lost distribution")

    # Retina!
    fig.set_dpi(600)

    return fig




"""Client implementation that only uses Uniswap v2 on-chain data to generate datasets."""
import logging
from types import NoneType
from typing import Tuple, cast

from pyarrow import Table

from web3 import Web3, HTTPProvider
from eth_defi.abi import get_contract
from eth_defi.event_reader.conversion import decode_data, convert_uint256_bytes_to_address, convert_int256_bytes_to_int, convert_uint256_hex_string_to_address
from eth_defi.event_reader.filter import Filter
from eth_defi.event_reader.logresult import LogResult
from eth_defi.event_reader.reader import read_events
from eth_defi.token import fetch_erc20_details
from eth_typing import HexAddress, HexStr

from tradingstrategy.chain import ChainId
from tradingstrategy.exchange import Exchange, ExchangeType, ExchangeUniverse
from tradingstrategy.pair import DEXPair, PandasPairUniverse
from tradingstrategy.stablecoin import is_stablecoin_like
from tradingstrategy.testing.mock_client import MockClient


logger = logging.getLogger(__name__)


class UniswapV2MockClient(MockClient):
    """A mock client that reads data from a single Uniswap v2 exchange directly from the chain.

    Designed to run tests against test EVM backends where we cannot generate
    proper test data because of the backends being temporary. This way we can skip the ETL
    step and pretend that the data is just there, but still have meaningful interaction
    with trading strategies with pairs and trade execution and such.

    Currently supported

    - Exchanges

    - Pairs data

    Any data is read from the chain on construction, then cached for subsequent fetch calls.

    .. warning::

        This client is not suitable to iterate real on-chain data.
        Due to high amount of real pairs deployed, you will need to wait
        several hours for :py:methd:`read_onchain_data` to complete.

    """

    def __init__(self,
                 web3: Web3,
                factory_address: HexAddress | str,
                router_address: HexAddress | str,
                init_code_hash: HexStr | str | NoneType = None,
                fee: float = 0.0030,
                 ):

        assert factory_address is not None, "factory_address not set"
        assert router_address is not None, "router_address not set"
        assert init_code_hash is not None, "init_code_hash not set"

        self.exchange_universe, self.pairs_table = UniswapV2MockClient.read_onchain_data(
            web3,
            factory_address,
            router_address,
            init_code_hash,
            fee,
        )
        assert len(self.pairs_table) > 0, f"Could not read any pairs from on-chain data. Uniswap v2 factory: {factory_address}, router: {router_address}."

    def get_default_quote_token_address(self) -> str:
        """Get the quote token address used in the generated pair map.

        Helper method for setting up simple local dev routing.
        """
        quote_tokens = set()
        pairs_df = self.fetch_pair_universe().to_pandas()
        pair_universe = PandasPairUniverse(pairs_df)
        assert pair_universe.get_count() > 0, "Pair universe has no trading pairs"
        for pair in pair_universe.iterate_pairs():
            quote_tokens.add(pair.quote_token_address)
        assert len(quote_tokens) == 1, f"Got {len(quote_tokens)} quote tokens in the pair universe, the pair universe is total {pair_universe.get_count()} pairs"
        return next(iter(quote_tokens))

    def fetch_exchange_universe(self) -> ExchangeUniverse:
        return self.exchange_universe

    def fetch_pair_universe(self) -> Table:
        return self.pairs_table

    @staticmethod
    def read_onchain_data(
                web3: Web3,
                factory_address: HexAddress | str,
                router_address: HexAddress | str,
                init_code_hash: HexStr | str | NoneType = None,
                fee: float = 0.0030,
    ) -> Tuple[ExchangeUniverse, Table]:
        """Reads Uniswap v2 data from EVM backend and creates tables for it.

        - Read data from a single Uniswap v2 compatible deployment

        - Read all PairCreated events and constructed Pandas DataFrame out of them

        :param fee:
            Uniswap v2 do not have fee information available on-chain, so we need to pass it.

            Default to 30 BPS.
        """

        chain_id = ChainId(web3.eth.chain_id)

        # Get contracts
        Factory = get_contract(web3, "sushi/UniswapV2Factory.json")

        start_block = 1
        end_block = web3.eth.block_number

        if isinstance(web3.provider, HTTPProvider):
            endpoint_uri = web3.provider.endpoint_uri
        else:
            endpoint_uri = str(web3.provider)

        # Assume logging is safe, because this mock client is only to be used with testing backends
        logger.info("Scanning PairCreated events, %d - %d, from %s, factory is %s",
                    start_block,
                    end_block,
                    endpoint_uri,
                    factory_address
                    )

        filter = Filter.create_filter(
            factory_address,
            [Factory.events.PairCreated],
        )

        # Read through all the events, all the chain, using a single threaded slow loop.
        # Only suitable for test EVM backends.
        pairs = []
        log: LogResult
        for log in read_events(
            web3,
            start_block,
            end_block,
            filter=filter,
            extract_timestamps=None,
        ):
            # Signature this
            #
            #  event PairCreated(address indexed token0, address indexed token1, address pair, uint);
            #
            # topic 0 = keccak(event signature)
            # topic 1 = token 0
            # topic 2 = token 1
            # argument 0 = pair
            # argument 1 = pair id
            #
            # log for EthereumTester backend is
            #
            # {'type': 'mined',
            #  'logIndex': 0,
            #  'transactionIndex': 0,
            #  'transactionHash': HexBytes('0x2cf4563f8c275e5b5d7a4e5496bfbaf15cc00d530f15f730ac4a0decbc01d963'),
            #  'blockHash': HexBytes('0x7c0c6363bc8f4eac452a37e45248a720ff09f330117cdfac67640d31d140dc38'),
            #  'blockNumber': 6,
            #  'address': '0xF2E246BB76DF876Cef8b38ae84130F4F55De395b',
            #  'data': HexBytes('0x00000000000000000000000068931307edcb44c3389c507dab8d5d64d242e58f0000000000000000000000000000000000000000000000000000000000000001'),
            #  'topics': [HexBytes('0x0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9'),
            #   HexBytes('0x0000000000000000000000002946259e0334f33a064106302415ad3391bed384'),
            #   HexBytes('0x000000000000000000000000b9816fc57977d5a786e654c7cf76767be63b966e')],
            #  'context': None,
            #  'event': web3._utils.datatypes.PairCreated,
            #  'chunk_id': 1,
            #  'timestamp': None}
            #
            arguments = decode_data(log["data"])
            topics = log["topics"]
            token0 = convert_uint256_hex_string_to_address(topics[1])
            token1 = convert_uint256_hex_string_to_address(topics[2])
            pair_address = convert_uint256_bytes_to_address(arguments[0])
            pair_id = convert_int256_bytes_to_int(arguments[1])

            token0_details = fetch_erc20_details(web3, token0)
            token1_details = fetch_erc20_details(web3, token1)

            # Our very primitive check to determine base token and quote token.
            # It's ok because this is a test backend
            if is_stablecoin_like(token0_details.symbol):
                quote_token_details = token0_details
                base_token_details = token1_details
            elif is_stablecoin_like(token1_details.symbol):
                quote_token_details = token1_details
                base_token_details = token0_details
            else:
                raise NotImplementedError(f"Does not know how to handle base-quote pairing for {token0_details} - {token1_details}. You need to update UniswapV2MockClient with more logic.")

            pair = DEXPair(
                    pair_id=pair_id,
                    chain_id=chain_id,
                    exchange_id=1,
                    exchange_slug="UniswapV2MockClient",
                    exchange_address=factory_address.lower(),
                    pair_slug=f"{base_token_details.symbol.lower()}-{quote_token_details.symbol.lower()}",
                    address=pair_address.lower(),
                    dex_type=ExchangeType.uniswap_v2,
                    base_token_symbol=base_token_details.symbol,
                    quote_token_symbol=quote_token_details.symbol,
                    token0_decimals=token0_details.decimals,
                    token1_decimals=token1_details.decimals,
                    token0_symbol=token0_details.symbol,
                    token1_symbol=token1_details.symbol,
                    token0_address=token0.lower(),
                    token1_address=token1.lower(),
                    buy_tax=0,
                    sell_tax=0,
                    transfer_tax=0,
                    fee=int(fee * 10_000),  #  Convert to BPS
                )
            pairs.append(pair)

        exchange = Exchange(
            chain_id=chain_id,
            chain_slug=chain_id.get_slug(),
            exchange_slug="UniswapV2MockClient",
            exchange_id=1,
            address=factory_address.lower(),
            exchange_type=ExchangeType.uniswap_v2,
            pair_count=len(pairs),
            default_router_address=router_address.lower(),
            init_code_hash=init_code_hash,
        )

        exchange_universe = ExchangeUniverse(exchanges={1: exchange})

        pair_table = DEXPair.convert_to_pyarrow_table(pairs, check_schema=False)

        return exchange_universe, pair_table








import datetime
from typing import Set

from requests import HTTPError, Timeout, TooManyRedirects

from tradingstrategy.client import Client
from tradingstrategy.timebucket import TimeBucket
from tradingstrategy.types import PrimaryKey

import abc
import datetime
import time
from dataclasses import dataclass
import logging
from typing import Set

from tradingstrategy.client import Client
from tradingstrategy.timebucket import TimeBucket
from tradingstrategy.types import PrimaryKey


logger = logging.getLogger(__name__)


#: Errors that are likely caused by flaky internet connection
#: and we should retry
RETRYABLE_EXCEPTIONS = (ConnectionError, HTTPError, Timeout, TooManyRedirects)


def wait_for_data(client: Client,
                  pair_ids: Set[PrimaryKey],
                  tick: datetime.datetime,
                  bucket: TimeBucket,
                  maximum_wait: datetime.timedelta,
                  sleep: datetime.timedelta = datetime.timedelta(seconds=5),
                  retryable_exceptions=RETRYABLE_EXCEPTIONS,
                  ):
    """Wait until the candle data is available.

    We wait until we have the data for the previous candle
    available.
    """

    logger.info("Waiting for strategy data for tick: %s, candle: %s")

    deadline = datetime.datetime.utcnow() + maximum_wait

    # Query only the latest candle from JSONL endpoint
    candle_ts = tick - bucket.to_pandas_timedelta()
    start_ts = candle_ts - datetime.timedelta(seconds=1)
    end_ts = candle_ts

    attempts = 1

    while datetime.datetime.utcnow() < deadline:
        logger.info("Starting to check the data age, attempt #%d", attempts)

        try:
            candles = client.fetch_candles_by_pair_ids(
                pair_ids,
                bucket,
                start_ts,
                end_ts,
            )

        except Exception as e:
            if not isinstance(e, retryable_exceptions):
                raise RuntimeError(f"Data wait fetch loop aborted") from e

        time.sleep(sleep.total_seconds())

        attempts += 1



"""Analyze the trade performance of algorithm

.. warning ::

    This module is deprecated and replaced by tradeexecutor analysis modules.

- Trade summary
- Won and lost trades
- Trade won/lost distribution graph
- Trade timeline and analysis of individual trades made

"""
import datetime
from dataclasses import dataclass, field
from typing import List, Dict, Iterable, Optional, Tuple, Callable

import numpy as np
import pandas as pd
from pandas.io.formats.style import Styler

from tradingstrategy.analysis.tradehint import TradeHint, TradeHintType
from tradingstrategy.exchange import ExchangeUniverse
from tradingstrategy.pair import LegacyPairUniverse, PandasPairUniverse
from tradingstrategy.types import PrimaryKey, USDollarAmount
from tradingstrategy.utils.format import format_value, format_percent, format_price, format_duration_days_hours_mins, \
    format_percent_2_decimals
from tradingstrategy.utils.summarydataframe import as_dollar, as_integer, create_summary_table, as_percent, as_missing


@dataclass
class SpotTrade:
    """Track spot trades to construct position performance.

    For sells, quantity is negative.
    """

    #: Internal running counter to uniquely label all trades in trade analysis
    trade_id: PrimaryKey

    #: Trading pair for this trade
    pair_id: PrimaryKey

    #: When this trade was made, the backtes simulation thick
    timestamp: pd.Timestamp

    #: Asset price at buy in
    price: USDollarAmount

    #: How much we bought the asset. Negative value for sells.
    quantity: float

    #: How much fees we paid to the exchange
    commission: USDollarAmount

    #: How much we lost against the midprice due to the slippage
    slippage: USDollarAmount

    #: Any hints applied for this trade why it was performed
    hint: Optional[TradeHint] = None

    #: Internal state dump of the algorithm when this trade was made.
    #: This is mostly useful when doing the trade analysis try to understand
    #: why some trades were made.
    #: It also allows you to reconstruct the portfolio state over the time.
    state_details: Optional[Dict] = None

    def is_buy(self):
        return self.quantity > 0

    def is_sell(self):
        return self.quantity < 0

    @property
    def value(self) -> USDollarAmount:
        return  abs(self.price * self.quantity)


@dataclass
class TradePosition:
    """How a particular asset traded.

    Each asset can have multiple entries (buys) and exits (sells)

    For a simple strategies there can be only one or two trades per position.

    * Enter (buy)

    * Exit (sell optionally)
    """

    #: List of all trades done for this position
    trades: List[SpotTrade] = field(default_factory=list)

    #: Closing the position could be deducted from the trades themselves,
    #: but we cache it by hand to speed up processing
    opened_at: Optional[pd.Timestamp] = None

    #: Closing the position could be deducted from the trades themselves,
    #: but we cache it by hand to speed up processing
    closed_at: Optional[pd.Timestamp] = None

    def __eq__(self, other: "TradePosition"):
        """Trade positions are unique by opening timestamp and pair id.]

        We assume there cannot be a position opened for the same asset at the same time twice.
        """
        return self.position_id == other.position_id

    def __hash__(self):
        """Allows easily create index (hash map) of all positions"""
        return hash((self.position_id))

    @property
    def position_id(self) -> PrimaryKey:
        """Position id is the same as the opening trade id."""
        return self.trades[0].trade_id

    @property
    def pair_id(self) -> PrimaryKey:
        """Position id is the same as the opening trade id."""
        return self.trades[0].pair_id

    @property
    def duration(self) -> Optional[datetime.timedelta]:
        """How long this position was held.

        :return: None if the position is still open
        """
        if not self.is_closed():
            return None
        return self.closed_at - self.opened_at

    def is_open(self):
        return self.closed_at is None

    def is_closed(self):
        return not self.is_open()

    @property
    def open_quantity(self) -> float:
        return sum([t.quantity for t in self.trades])

    @property
    def open_value(self) -> float:
        """The current value of this open position, with the price at the time of opening."""
        assert self.is_open()
        return sum([t.value for t in self.trades])

    @property
    def open_price(self) -> float:
        """At what price we opened this position.

        Supports only simple enter/exit positions.
        """
        return self.get_first_entry_price()

    def get_first_entry_price(self) -> float:
        """What was the price when the first entry buy for this position was made.
        """
        buys = list(self.buys)
        return buys[0].price

    def get_last_exit_price(self) -> float:
        """What was the time when the last sell for this position was executd.
        """
        sells = list(self.sells)
        return sells[-1].price
        assert len(sells) == 1

    @property
    def close_price(self) -> float:
        """At what price we exited this position.

        Supports only simple enter/exit positions.
        """
        return self.get_last_exit_price()

    @property
    def buys(self) -> Iterable[SpotTrade]:
        return [t for t in self.trades if t.is_buy()]

    @property
    def sells(self) -> Iterable[SpotTrade]:
        return [t for t in self.trades if t.is_sell()]

    @property
    def buy_value(self) -> USDollarAmount:
        return sum([t.value - t.commission for t in self.trades if t.is_buy()])

    @property
    def sell_value(self) -> USDollarAmount:
        return sum([t.value - t.commission for t in self.trades if t.is_sell()])

    @property
    def realised_profit(self) -> USDollarAmount:
        """Calculated life-time profit over this position."""
        assert not self.is_open()
        return -sum([t.quantity * t.price - t.commission for t in self.trades])

    @property
    def realised_profit_percent(self) -> float:
        """Calculated life-time profit over this position."""
        assert not self.is_open()
        buy_value = self.buy_value
        sell_value = self.sell_value
        return sell_value / buy_value - 1

    def is_win(self):
        """Did we win this trade."""
        assert not self.is_open()
        return self.realised_profit > 0

    def is_lose(self):
        assert not self.is_open()
        return self.realised_profit < 0

    def is_stop_loss(self) -> bool:
        """Was stop loss triggered for this position"""
        for t in self.trades:
            if t.hint:
                if t.hint.type == TradeHintType.stop_loss_triggered:
                    return True
        return False

    def add_trade(self, t: SpotTrade):
        if self.trades:
            last_trade = self.trades[-1]
            assert t.timestamp > last_trade.timestamp, f"Tried to do trades in wrong order. Last: {last_trade}, got {t}"
        self.trades.append(t)

    def can_trade_close_position(self, t: SpotTrade):
        assert self.is_open()
        if not t.is_sell():
            return False
        open_quantity = self.open_quantity
        closing_quantity = -t.quantity
        assert closing_quantity <= open_quantity, "Cannot sell more than we have in balance sheet"
        return closing_quantity == open_quantity

    def get_max_size(self) -> USDollarAmount:
        """Get the largest size of this position over the time"""
        cur_size = 0
        max_size = 0
        for t in self.trades:
            cur_size += t.value
            max_size = max(cur_size, max_size)
        return max_size

    def get_trade_count(self) -> int:
        """How many individual trades was done to manage this position."""
        return len(self.trades)


@dataclass
class AssetTradeHistory:
    """How a particular asset traded.

    Each position can have increments or decrements.
    When position is decreased to zero, it is considered closed, and a new buy open a new position.
    """
    positions: List[TradePosition] = field(default_factory=list)

    def get_first_opened_at(self) -> Optional[pd.Timestamp]:
        if self.positions:
            return self.positions[0].opened_at
        return None

    def get_last_closed_at(self) -> Optional[pd.Timestamp]:
        for position in reversed(self.positions):
            if not position.is_open():
                return position.closed_at

        return None

    def add_trade(self, t: SpotTrade):
        """Adds a new trade to the asset history.

        If there is an open position the trade is added against this,
        otherwise a new position is opened for tracking.
        """
        current_position = None
        if self.positions:
            if self.positions[-1].is_open():
                current_position = self.positions[-1]

        if current_position:
            if current_position.can_trade_close_position(t):
                # Close the existing position
                current_position.closed_at = t.timestamp
                current_position.add_trade(t)
                assert current_position.open_quantity == 0
            else:
                # Add to the existing position
                current_position.add_trade(t)
        else:
            # Open new position
            new_position = TradePosition(opened_at=t.timestamp)
            new_position.add_trade(t)
            self.positions.append(new_position)


@dataclass
class TradeSummary:
    """Some generic statistics over all the trades"""
    won: int
    lost: int
    zero_loss: int
    stop_losses: int
    undecided: int
    realised_profit: USDollarAmount
    open_value: USDollarAmount
    uninvested_cash: USDollarAmount
    initial_cash: USDollarAmount
    extra_return: USDollarAmount

    def to_dataframe(self) -> pd.DataFrame:
        """Creates a human-readable Pandas dataframe table from the object."""
        total_trades = self.won + self.lost
        human_data = {
            "Return %": as_percent(self.realised_profit / self.initial_cash),
            "Cash at start": as_dollar(self.initial_cash),
            "Value at end": as_dollar(self.open_value + self.uninvested_cash),
            "Trade win percent": as_percent(self.won / total_trades) if total_trades else as_missing(),
            "Total trades done": as_integer(self.won + self.lost + self.zero_loss),
            "Won trades": as_integer(self.won),
            "Lost trades": as_integer(self.lost),
            "Stop losses triggered": as_integer(self.stop_losses),
            "Stop loss % of all": as_percent(self.stop_losses / total_trades) if total_trades else as_missing(),
            "Stop loss % of lost": as_percent(self.stop_losses / self.lost) if self.lost else as_missing(),
            "Zero profit trades": as_integer(self.zero_loss),
            "Positions open at the end": as_integer(self.undecided),
            "Realised profit and loss": as_dollar(self.realised_profit),
            "Portfolio unrealised value": as_dollar(self.open_value),
            "Extra returns on lending pool interest": as_dollar(self.extra_return),
            "Cash left at the end": as_dollar(self.uninvested_cash),
        }
        return create_summary_table(human_data)


@dataclass
class TradeAnalyzer:
    """Analysis of trades in a portfolio."""

    #: How a particular asset traded. Asset id -> Asset history mapping
    asset_histories: Dict[object, AssetTradeHistory] = field(default_factory=dict)

    def get_first_opened_at(self) -> Optional[pd.Timestamp]:
        def all_opens():
            for history in self.asset_histories.values():
                yield history.get_first_opened_at()

        return min(all_opens())

    def get_last_closed_at(self) -> Optional[pd.Timestamp]:
        def all_closes():
            for history in self.asset_histories.values():
                closed = history.get_last_closed_at()
                if closed:
                    yield closed
        return max(all_closes())

    def get_all_positions(self) -> Iterable[Tuple[PrimaryKey, TradePosition]]:
        """Return open and closed positions over all traded assets."""
        for pair_id, history in self.asset_histories.items():
            for position in history.positions:
                yield pair_id, position

    def get_open_positions(self) -> Iterable[Tuple[PrimaryKey, TradePosition]]:
        """Return open and closed positions over all traded assets."""
        for pair_id, history in self.asset_histories.items():
            for position in history.positions:
                if position.is_open():
                    yield pair_id, position

    def calculate_summary_statistics(self, initial_cash, uninvested_cash, extra_return=0) -> TradeSummary:
        """Calculate some statistics how our trades went."""
        won = lost = zero_loss = stop_losses = undecided = 0
        open_value: USDollarAmount = 0
        profit: USDollarAmount = 0
        for pair_id, position in self.get_all_positions():
            if position.is_open():
                open_value += position.open_value
                undecided += 1
                continue

            if position.is_stop_loss():
                stop_losses += 1

            if position.is_win():
                won += 1
            elif position.is_lose():
                lost += 1
            else:
                # Any profit exactly balances out loss in slippage and commission
                zero_loss += 1

            profit += position.realised_profit

        return TradeSummary(
            won=won,
            lost=lost,
            zero_loss=zero_loss,
            stop_losses=stop_losses,
            undecided=undecided,
            realised_profit=profit + extra_return,
            open_value=open_value,
            uninvested_cash=uninvested_cash,
            initial_cash=initial_cash,
            extra_return=extra_return,
        )

    def create_timeline(self) -> pd.DataFrame:
        """Create a timeline feed how we traded over a course of time.

        Note: We assume each position has only one enter and exit event, not position increases over the lifetime.

        :return: DataFrame with timestamp and timeline_event columns
        """

        def gen_events():
            for pair_id, position in self.get_all_positions():
                yield (position.position_id, position)

        df = pd.DataFrame(gen_events(), columns=["position_id", "position"])
        return df


def expand_timeline(
        exchange_universe: ExchangeUniverse,
        pair_universe: PandasPairUniverse,
        timeline: pd.DataFrame,
        vmin=-0.3,
        vmax=0.2,
        timestamp_format="%Y-%m-%d",
        hidden_columns=["Id", "PnL % raw"]) -> Tuple[pd.DataFrame, Callable]:
    """Expand trade history timeline to human readable table.

    This will the outputting much easier in Python Notebooks.

    Currently does not incrementing/decreasing positions gradually.

    Instaqd of applying styles or returning a styled dataframe, we return a callable that applies the styles.
    This is because of Pandas issue https://github.com/pandas-dev/pandas/issues/40675 - hidden indexes, columns,
    etc. are not exported.

    :param vmax: Trade success % to have the extreme green color.

    :param vmin: The % of lost capital on the trade to have the extreme red color.

    :param timestamp_format: How to format Opened at column, as passed to `strftime()`

    :param hidden_columns: Hide columns in the output table

    :return: DataFrame with human readable position win/loss information, having DF indexed by timestamps and a styler function
    """

    # https://stackoverflow.com/a/52363890/315168
    def expander(row):
        position: TradePosition = row["position"]
        # timestamp = row.name  # ???
        pair_id = position.pair_id
        pair_info = pair_universe.get_pair_by_id(pair_id)
        exchange = exchange_universe.get_by_id(pair_info.exchange_id)

        remarks = "SL" if position.is_stop_loss() else ""

        r = {
            # "timestamp": timestamp,
            "Id": position.position_id,
            "Remarks": remarks,
            "Opened at": position.opened_at.strftime(timestamp_format),
            "Duration": format_duration_days_hours_mins(position.duration) if position.duration else np.nan,
            "Exchange": exchange.name,
            "Base asset": pair_info.base_token_symbol,
            "Quote asset": pair_info.quote_token_symbol,
            "Position max size": format_value(position.get_max_size()),
            "PnL USD": format_value(position.realised_profit) if position.is_closed() else np.nan,
            "PnL %": format_percent_2_decimals(position.realised_profit_percent) if position.is_closed() else np.nan,
            "PnL % raw": position.realised_profit_percent if position.is_closed() else 0,
            "Open price USD": format_price(position.open_price),
            "Close price USD": format_price(position.close_price) if position.is_closed() else np.nan,
            "Trade count": position.get_trade_count(),
        }
        return r

    applied_df = timeline.apply(expander, axis='columns', result_type='expand')

    # https://stackoverflow.com/a/52720936/315168
    applied_df\
        .sort_values(by=['Id'], ascending=[True], inplace=True)

    # Get rid of NaN labels
    # https://stackoverflow.com/a/28390992/315168
    applied_df.fillna('', inplace=True)

    def apply_styles(df: pd.DataFrame):
        # Create a Pandas Styler with multiple styling options applied
        # Dynamically color the background of trade outcome coluns # https://pandas.pydata.org/docs/reference/api/pandas.io.formats.style.Styler.background_gradient.html
        styles = df.style\
            .hide_index()\
            .hide_columns(hidden_columns)\
            .background_gradient(
                axis=0,
                gmap=applied_df['PnL % raw'],
                cmap='RdYlGn',
                vmin=vmin,  # We can only lose 100% of our money on position
                vmax=vmax)  # 50% profit is 21.5 position. Assume this is the max success color we can hit over

        # Don't let the text inside a cell to wrap
        styles = styles.set_table_styles({
            "Opened at": [{'selector': 'td', 'props': [('white-space', 'nowrap')]}],
            "Exchange": [{'selector': 'td', 'props': [('white-space', 'nowrap')]}],
        })
        return styles

    return applied_df, apply_styles



"""Hint trade types.

.. warning ::

    This module is deprecated and replaced by tradeexecutor analysis modules.

"""

import enum
from typing import Optional


class TradeHintType(enum.Enum):
    """Add hints to the backtested trades.

    This help later to analyse the backtested trades and differ e.g. stop losses from normal position closes.
    """

    open = "open"
    close = "close"
    stop_loss_triggered = "stop_loss_triggered"


class TradeHint:

    def __init__(self, type: TradeHintType, message: Optional[str]=None):
        self.type = type
        self.message = message


"""Drawing OHLCV candle charts using Plotly.

Draw price charts using Plotly.

- Candle charts

- Open, high, low, close (OHLC)

- Open, high, low, close, volume (OHLCV)

- Dark and light themes
"""
import enum
import logging
from typing import Optional

import plotly.graph_objects as go
import pandas as pd
from plotly.subplots import make_subplots


logger = logging.getLogger(__name__)


class BadOHLCVData(Exception):
    """We could not figure out the data frame"""


class VolumeBarMode(enum.Enum):
    """Should candlestick chart come with the volume bars."""

    #: Volume bars are in a chart below candle chart
    separate = "separate"

    #: Volume bars are transparently inside the candle chart
    overlay = "overlay"

    #: Do not show volume bars
    hidden = "hidden"


def validate_ohclv_dataframe(candles: pd.DataFrame):
    """Ensures that ohclv dataframe contains valid data:
    1. Must either contain an index or column named date or timestamp
    2. Must contain columns named open, close, high, and low"""

    required_columns = ["open", "close", "high", "low"]

    if (
        candles.index.name not in {"date", "timestamp"} and \
        not ({"date", "timestamp"}).issubset(candles.columns)
    ):
        raise BadOHLCVData("OHLCV DataFrame lacks date/timestamp index or column")

    for r in required_columns:
        if r not in candles.columns:
            raise BadOHLCVData(f"OHLCV DataFrame lacks column: {r}, has {candles.columns}")


def create_label(row: pd.Series) -> str:
    """Create labels for a single candle."""


def make_candle_labels(
        df: pd.DataFrame,
        dollar_prices=True,
        base_token_name: Optional[str]=None,
        quote_token_name: Optional[str]=None,
        line_separator="<br>",
        candle_decimals: int=4,
) -> pd.Series:
    """Generate individual labels for OHLCV chart candles.

    Used to display toolips on OHLCV chart.

    :poram candle_df:
        Candles for which we need tooltips.

    :poram label_df:
        A target dataframe

        A column "label" is generated and it is populated
        for every index timestamp that does not have label yet.

    :param dollar_prices:
        True if prices are in USD. Otherwise in the given quote token.

    :param quote_token_name:
        Cryptocurrency as the quote token pair.

    :param line_separator:
        New line format.

        Plotly wants raw HTML.

    :return:
        Series of text label
    """

    validate_ohclv_dataframe(df)

    if dollar_prices:
        if quote_token_name:
            price_text = f"{quote_token_name}/USD"
            volume_text = "USD"
        else:
            price_text = "USD"
            volume_text = "USD"
    else:
        assert quote_token_name, "Quote token must be given"
        price_text = f"{base_token_name} / {quote_token_name}"
        volume_text = quote_token_name

    # All label values are NA by default
    def _create_label_for_single_candle(row: pd.Series):
        # Index here can be MultiIndex as well, so assume timestamp is available as a column

        timestamp = row["timestamp"]

        percentage_change = (row.close - row.open) / row.open * 100
        text = [
            f"{timestamp.strftime('%Y-%m-%d %H:%M:%S UTC')}",
            "",
            f"Open: {round(row.open, candle_decimals)} {price_text}",
            f"High: {round(row.high, candle_decimals)} {price_text}",
            f"Low: {round(row.low, candle_decimals)} {price_text}",
            f"Close: {round(row.close, candle_decimals)} {price_text}",
        ]

        if "volume" in row.index:
            text += [
                f"Volume: {round(row.volume, candle_decimals)} {volume_text}",
            ]

        text += [
            f"Change: {percentage_change:.2f} %",
            "",
        ]

        if "exchange_rate" in row.index:
            text += [f"Exchange rate: {row.exchange_rate} {quote_token_name} / USD", ""]

        if "buys" in row.index:
            text += [
                f"Buys: {row.buys} txs",
                f"Sells: {row.sells} txs",
                f"Total: {row.buys + row.sells} trades",
                ""
            ]

        return line_separator.join(text)

    return df.apply(_create_label_for_single_candle, axis="columns")


def visualise_ohlcv(
        candles: pd.DataFrame,
        chart_name: Optional[str] = None,
        y_axis_name: Optional[str] = "Price USD",
        volume_axis_name: Optional[str] = "Volume USD",
        height: int = 800,
        theme: str = "plotly_white",
        volume_bar_colour: str = "rgba(128,128,128,0.5)",
        volume_bar_mode = VolumeBarMode.overlay,
        labels: Optional[pd.Series] = None,
        num_detached_indicators: Optional[int] = 0,
        vertical_spacing: Optional[float] = 0.05,
        relative_sizing: Optional[list[float]]  = None,
        subplot_names: Optional[list[str]] = None,
        subplot_font_size: int = 11,
) -> go.Figure:
    """Draw a candlestick chart.

    If the `candles` has `label` column this will be used
    as the mouse hover text for candles.

    If the `candles` has `volume` column, draw also this column.

    We remove the default "minimap" scrolling as it has pretty
    bad usability.

    :param chart_name:
        Will be displayed at the top of the chart

    :param y_axis_name:
        Will be displayed on an Y-axis

    :param height:
        Chart height in pixels

    :param theme:
        Plotly colour scheme for the chart.

        `See Plotly color scheme list here <https://plotly.com/python/templates/>`__.

    :param volume_bar_colour:
        Override the default colour for volume bars

    :param volume_bar_mode:

        Draw volume chart as a separate chart under the candlestick chart.

        If not set, draw as an overlay.

        `Note that Plotly does not allow reodering of tracing <https://github.com/plotly/plotly.py/issues/2345#issuecomment-809339043>`__,
        and the volume bars will always be on the top of the candlesticks.

    :param labels:
        Tooltip labels for candles.

        See :py:func:`make_candle_labels`
    
    :param num_detached_indicators:
        Number of indicators that will be drawn as separate charts. Includes volume if volume is separate.
    
    :param vertical_spacing:
        Vertical spacing between charts.
    
    :param relative_sizing:
        Sizing of subplots relative to the main price chart. 
        Price chart is regarded as 1.0, so subplots should be smaller than 1.0.
        Should include size for volume subplot if volume is separate.
        
        If it is overlayed, volume will naturally be the same size as the price chart, since they're on the same chart.
    
    :param subplot_names:
        Names of subplots. Used as titles for subplots. 
        
        Should include name for main candle chart. Recommended to leave candle chart name as "" or None
        
        Should include volume if volume is separate.
        
    :return:
        Plotly figure object
    """

    # sanity checks
    
    validate_plot_info(volume_bar_mode, num_detached_indicators, relative_sizing, subplot_names)
    
    validate_ohclv_dataframe(candles)

    if labels is not None:
        text = labels
    else:
        # TODO: Legacy - deprecate
        # Add change percentages on candle mouse hover
        percentage_changes = ((candles['close'] - candles['open'])/candles['open']) * 100
        text = ["Change: " + f"{percentage_changes[i]:.2f}%" for i in range(len(percentage_changes))]

    candlesticks = go.Candlestick(
        x=candles.index,
        open=candles['open'],
        high=candles['high'],
        low=candles['low'],
        close=candles['close'],
        showlegend=False,
        text=text,
        hoverinfo="text",
    )

    
    if "volume" in candles.columns:
        volume_bars = go.Bar(
                x=candles.index,
                y=candles['volume'],
                showlegend=False,
                marker={
                    "color": volume_bar_colour,
                }
            )
    else:
        volume_bars = None
    
    # Add volume to plot
    fig = _get_volume_grid(
        volume_bars, 
        volume_bar_mode, 
        volume_axis_name, 
        num_detached_indicators,
        vertical_spacing,
        relative_sizing,
        subplot_names,
    )
    
    # Set chart core options
    _set_chart_core_options(chart_name, y_axis_name, height, theme, fig)
    
    # Add candlesticks last since we want them on top if overlayed
    fig.add_trace(candlesticks, secondary_y=False)
    
    fig.update_annotations(font_size=subplot_font_size)

    return fig

def validate_plot_info(
    volume_bar_mode: VolumeBarMode, 
    num_detached_indicators: int, 
    relative_sizing: list[float], 
    subplot_names: list[str],
) -> None | ValueError | AssertionError:
    """Validate plot info to ensure that it is valid for the given volume_bar_mode, num_detached_indicators, relative_sizing, and subplot_names"""
    
    if volume_bar_mode == VolumeBarMode.separate:
        assert num_detached_indicators > 0, "num_detached_indicators must be greater than 0 if volume_bar_mode is separate"
    elif volume_bar_mode in {VolumeBarMode.overlay, VolumeBarMode.hidden}:
        assert num_detached_indicators >= 0, "num_detached_indicators must be greater than or equal to 0 if volume_bar_mode is overlay or hidden"
    else:
        raise ValueError(f"Invalid volume_bar_mode. Got {volume_bar_mode}")
    
    tail = "\nAlso remember to include element for main price chart."
    
    if relative_sizing:
        _error_message = f"len(relative_sizing) ({len(relative_sizing)}) must be 1 greater than num_detached_indicators ({num_detached_indicators})"
        
        
        # add helpful error messages
        if volume_bar_mode == VolumeBarMode.separate:
            error_message = _error_message + "\nRemember to include volume subplot size since it is separate." + tail
        elif volume_bar_mode in {VolumeBarMode.overlay, VolumeBarMode.hidden}: 
            error_message = _error_message + "\nRemember to exclude volume subplot size since it is not separate." + tail
        else:
            raise ValueError(f"Invalid volume_bar_mode. Got {volume_bar_mode}")
        
        assert len(relative_sizing) == num_detached_indicators + 1, error_message
        
    if subplot_names:
        _error_message = f"len(subplot_names) ({len(subplot_names)}) must be 1 greater than num_detached_indicators ({num_detached_indicators})"
        
        # add helpful error messages
        if volume_bar_mode == VolumeBarMode.separate:
            error_message = _error_message + "\nRemember to include volume subplot name since it is separate." + tail
        elif volume_bar_mode in {VolumeBarMode.overlay, VolumeBarMode.hidden}:
            error_message = _error_message + "\nRemember to exclude volume subplot name since it is not separate." + tail
        else:
            raise ValueError(f"Invalid volume_bar_mode. Got {volume_bar_mode}")
        
        assert len(subplot_names) == num_detached_indicators + 1, error_message        

def _set_chart_core_options(
    chart_name: str, 
    y_axis_name: str, 
    height: int, 
    theme: str, 
    fig: go.Figure,
):
    """Update figure layout. Set chart core options."""
    
    fig.update_layout(
        height=height,
        template=theme,
    )

    if chart_name:
        fig.update_layout(
            title=chart_name,
        )

    # Range slider is not very user friendly so just
    # disable it for now
    fig.update_xaxes(rangeslider={"visible": False})

    # Move legend to the bottom so we have more space for
    # time axis in narrow notebook views
    # https://plotly.com/python/legend/f
    # fig.update_layout(
    #     legend={
    #         "orientation": "h",
    #         "yanchor": "bottom",
    #         "y": 1.02,
    #         "xanchor": "right",
    #         "x": 1,
    #     })
    
    
    # moves legend to the top right (below the title and above the plotting area)
    # and makes it horizontal
    fig.update_layout(
        title=dict(
            y=0.95  # Position the title just below the top margin
        ),
        legend=dict(
            x=1,
            y=1.09,  # Position the legend just above the plot area
            xanchor='right',
            yanchor='top',
            orientation='h'  # Display the legend items horizontally
        ),
        margin=dict(t=150)  # Add some extra space at the top for the title and legend
    )

def _get_volume_grid(
    volume_bars: go.Bar, 
    volume_bar_mode: bool, 
    volume_axis_name: str, 
    num_detached_indicators: int,
    vertical_spacing: float,
    relative_sizing: list[float],
    subplot_names: list[str],
) -> go.Figure:
    """Get subplot grid, with volume information, based on the volume bar mode"""
    
    is_secondary_y = _get_secondary_y(volume_bar_mode)

    if relative_sizing and not all(relative_sizing):
        raise ValueError(
            f"relative_sizing must be a list of floats. Got: {relative_sizing}"
        )

    if volume_bar_mode == VolumeBarMode.separate:

        fig = _get_grid_without_volume(num_detached_indicators, vertical_spacing, relative_sizing, subplot_names, is_secondary_y)

        if volume_bars is not None:
            # https://stackoverflow.com/a/65997291/315168
            _update_separate_volume(volume_bars, volume_axis_name, fig)

        return fig

    elif volume_bar_mode == VolumeBarMode.overlay:
        
        fig = _get_grid_without_volume(num_detached_indicators, vertical_spacing, relative_sizing, subplot_names, is_secondary_y)

        if volume_bars is not None:
            # If overlayed, we need to add volume first
            _update_overlay_volume(volume_bars, volume_axis_name, fig)

        return fig

    elif volume_bar_mode == VolumeBarMode.hidden:
        # no need to add volume bars
        
        return _get_grid_without_volume(num_detached_indicators, vertical_spacing, relative_sizing, subplot_names, is_secondary_y)

    else:
        raise ValueError(f"Unknown volume bar mode: {volume_bar_mode}")

def _get_grid_without_volume(
    num_detached_indicators: int, 
    vertical_spacing: float, 
    relative_sizing: list[float],
    subplot_names: list[str],
    is_secondary_y: bool
):
    specs = _get_specs(num_detached_indicators, is_secondary_y)

    row_heights = _get_row_heights(num_detached_indicators, relative_sizing)

    fig = make_subplots(
            rows = num_detached_indicators + 1,
            cols = 1,
            specs=specs,
            shared_xaxes=True,
            row_heights=row_heights,
            vertical_spacing=vertical_spacing,
            row_titles=subplot_names,
        )
    
    return fig

def _get_specs(
    num_detached_indicators: int,
    is_secondary_y: bool
):
    specs = [[{}] for _ in range(num_detached_indicators)]
    specs.insert(0, [{"secondary_y": is_secondary_y}])
    return specs

def _get_row_heights(
    num_detached_indicators: int, 
    relative_sizing: list[float]
) -> list[float]:
    """Get a list of heights for each row in the subplot grid, including main candle chart"""
    return relative_sizing or [1] + [0.2 for _ in range(num_detached_indicators)]

def _update_overlay_volume(
    volume_bars: go.Bar,
    volume_axis_name: str,
    fig: go.Figure
):
    """Update overlay volume chart info"""
    fig.add_trace(volume_bars, secondary_y=True)
    fig.update_yaxes(secondary_y=True, showgrid=False)

    if volume_axis_name:
        fig.update_yaxes(title=volume_axis_name, secondary_y=True, row=1)

def _update_separate_volume(
    volume_bars: go.Bar, 
    volume_axis_name: str,
    fig: go.Figure
):
    """Update detached volume chart info"""
    fig.add_trace(volume_bars, row=2, col=1)

    # volume axis name added to subplot title for now (right side instead of left)
    # if volume_axis_name:
    #    fig.update_yaxes(title=volume_axis_name, row=2)

def _get_secondary_y(volume_mode: VolumeBarMode) -> bool:
    """Based on the volume bar mode, should we use secondary Y axis?
    
    Secondary data may not have volume available"""
    if volume_mode == VolumeBarMode.overlay:
        return True
    elif volume_mode in [VolumeBarMode.hidden, VolumeBarMode.separate]:
        return False
    else:
        raise ValueError(f"Unknown volume bar mode: {volume_mode}")
            



from typing import Optional

from tradingstrategy.environment.base import download_with_progress_plain
from tradingstrategy.environment.config import Configuration
from tradingstrategy.transport.cache import CachedHTTPTransport


def run_interactive_setup() -> Optional[Configuration]:
    """Do REPL interactive setup with the user in Jupyter notebook."""

    api_key = None
    transport = CachedHTTPTransport(download_with_progress_plain, api_key=None)

    # Check the server is up - will throw an exception in the case of an error
    transport.ping()

    has_api_key = None
    while has_api_key not in ("y", "n", ""):
        print("Using Trading Strategy requires you to have an API key.")
        has_api_key = input("Do you have an API key yet? [y/n]? ").lower()

    if has_api_key == "":
        print("Aborting setup")
        return None

    if has_api_key == "y":
        pass
    else:
        reply = input("Would you like to sign up on the mailing list and get a free API key now? [y/n]").lower()
        if reply != 'y':
            print("Please rerun choose 'y'` to the prompt next timew.")
            return None

        first_name = input("Your first name? ")
        last_name = input("Your last name? ")
        print("The API key will be delivered to your email inbox.")
        email = input("Your email address? ")

        resp = transport.register(first_name, last_name, email)
        if resp["status"] != "OK":
            print(f"Failed to register: {resp}")
            return None

        print(f"Signed up on the newsletter: {email} - please check your email for the API key")

    while True:
        print("The API key is in format 'secret-token:tradingstrategy-...'")
        api_key = input("Enter your API key from the welcome email you just received, including secret-token: part: ")

        if api_key.strip() == "" or len(api_key) < 10:
            raise AssertionError("Interactive setup aborted")

        if validate_api_key(api_key):
            break

    config = Configuration(api_key=api_key)

    print("The API key setup complete.")

    return config


def run_non_interactive_setup(**kwargs) -> Optional[Configuration]:
    api_key = kwargs.get("api_key")
    if not api_key:
        print("Can't setup config, please provide api_key argument")
        return None
    
    if not validate_api_key(api_key):
        print("Incorrect API key")
        return None

    config = Configuration(api_key=api_key)

    print("The API key setup complete.")

    return config


def validate_api_key(api_key: str) -> bool:
    # Watch out whitespace copy paste issues
    api_key = api_key.strip()  
    if api_key == "":
        print("Aborting setup")
        return False

    print(f"Testing out API key: {api_key[0:24]}")
    authenticated_transport = CachedHTTPTransport(download_with_progress_plain, api_key=api_key)
    try:
        welcome = authenticated_transport.message_of_the_day()
        print("The server replied accepted our API key and sent the following greetings:")
        print("Server version:", welcome["version"])
        print("Message of the day:", welcome["message"])
    except Exception as e:
        print(f"Received error: {e} - check your API key")
        return False

    return True   
